class Mysql

    # Result set.
    class Result
        include Enumerable

        attr_reader :fields

        # Create the next result object.
        def initialize( mysql, result )
            @mysql = mysql
            @result = result
            @num_rows = @num_fields = nil
            raise ArgumentError, "Invalid result object" if @result.nil? or @result.null?
            ObjectSpace.define_finalizer( self, Result.finalizer(@result) )
        end

        # Frees the result object.
        def free
            C::mysql_free_result(@result)
            @result = nil
            ObjectSpace.undefine_finalizer( self )
        end

        # @return [Integer] the number of rows in this result set
        def num_rows
            raise Error, "Result has been freed" unless @result
            @num_rows ||= C::mysql_num_rows(@result)
        end

        # @return [Integer] the number of columns in this result set
        def num_fields
            raise Error, "Result has been freed" unless @result
            @num_fields ||= C::mysql_num_fields(@result)
        end

        # @return [Array<Integer>] the array of number of chars for each column
        def fetch_lengths
            raise Error, "Result has been freed" unless @result
            lengths = C::mysql_fetch_lengths(@result)
            (lengths.nil? or lengths.null?) ? nil : lengths.read_array_of_long(num_fields)
        end

        # @return [Array<String>] Ary of elements of the next row.
        def fetch_row
            raise Error, "Result has been freed" unless @result
            row = C::mysql_fetch_row(@result)
            if row.nil? or row.null?
                nil
            else
                lengths = fetch_lengths
                ptrs = row.get_array_of_pointer(0, lengths.size)
                (0...lengths.size).map {|i|
                    ptr = ptrs[i]
                    ptr.nil? or ptr.null? ? nil : ptr.read_string(lengths[i])
                }
            end
        end

        # Iterates over all rows in this result set.
        # @yield [Array<String>] Called once for each row in this result set
        # @see fetch_row
        def each
            while row = fetch_row
                yield row
            end
        end

        # @return [Integer] the current position of the row cursor
        def row_tell
            raise Error, "Result has been freed" unless @result
            C::mysql_row_tell(@result)
        end

        # Sets the position of the row cursor.
        # @param [Integer] offset the new position of the row cursor
        # @return [Integer] the former position of the row cursor
        def row_seek( offset )
            raise Error, "Result has been freed" unless @result
            C::mysql_row_seek(@result, offset)
        end

        # @param [Boolean] with_table if true, fields are denoted with table "tablename.fieldname"
        # @return [Hash] hash of elements "field" => "value"
        def fetch_hash(with_table = false)
            return nil unless row = fetch_row
            keys = if with_table
                       @tblcolnames ||= fetch_fields.map{|f| "#{f.table}.#{f.name}"}
                   else
                       @colnames ||= fetch_fields.map{|f| f.name}
                   end

            hash = {}
            row.each_with_index do |value, i|
                hash[keys[i]] = value
            end
            hash
        end

        # Iterates over all rows in this result set.
        # @param [Boolean] with_table if true, fields are denoted with table "tablename.fieldname"
        # @yield [Hash] Called once for each row in this result set with a row-hash
        # @see fetch_hash
        def each_hash(with_table = false)
            while row = fetch_hash(with_table)
                yield row
            end
        end

        # @return [Integer] The position of the field cursor after the last fetch_field.
        def field_tell
            raise Error, "Result has been freed" unless @result
            C::mysql_field_tell(@result)
        end

        # Sets the field cursor to the given offset.
        # @param [Integer] the new field offset
        # @return [Integer] the position of the previous field cursor
        def field_seek(offset)
            raise Error, "Result has been freed" unless @result
            C::mysql_field_seek(@result, offset)
        end

        # @return [Field,nil] the information for the next Field or nil
        def fetch_field
            raise Error, "Result has been freed" unless @result
            field_ptr = C::mysql_fetch_field(@result)
            if field_ptr.nil? or field_ptr.null?
                nil
            else
                Field.new(C::Field.new(field_ptr))
            end
        end

        # @param [Integer] fieldnr column number
        # @return [Field] the information for field of column fieldnr
        def fetch_field_direct( fieldnr )
            raise Error, "Result has been freed" unless @result
            n = num_fields
            raise Error, "#{fieldnr}: out of range (max: #{n})" if fieldnr < 0 or fieldnr >= n
            field_ptr = C::mysql_fetch_field_direct(@result, fieldnr)
            Field.new(C::Field.new(field_ptr))
        end

        # @return [Array<Field>] array of field-informations for each column in this result set
        def fetch_fields
            raise Error, "Result has been freed" unless @result
            n = num_fields
            (0...n).map{|i| fetch_field_direct(i)}
        end

        # Seeks to an arbitrary row in the result set.
        # @param [Integer] row the number of row to use next
        # @return [self]
        def data_seek( row )
            raise Error, "Result has been freed" unless @result
            C::mysql_data_seek(@result, row)
            self
        end

        # Internal finalizer, calls self.free.
        def self.finalizer(result)
            Proc.new do |*args|
                C::mysql_free_result(result)
            end
        end
    end

end
